@@ -17,7 +17,7 @@ use LWP::UserAgent ();
 use JSON        ();
 use URI::Encode ();
 
-our $VERSION = '1.0002';
+our $VERSION = '1.0003';
 
 =encoding utf8
 
@@ -28,11 +28,14 @@ OpenStack::Client - A cute little client to OpenStack services
 =head1 SYNOPSIS
 
     #
-    # First, connect to an API endpoint via the Keystone authorization service
+    # First, connect to an API endpoint via the Keystone
+    # authorization service
     #
     use OpenStack::Client::Auth ();
 
-    my $auth = OpenStack::Client::Auth->new('http://openstack.foo.bar:5000/v2.0',
+    my $endpoint = 'http://openstack.foo.bar:5000/v2.0';
+
+    my $auth = OpenStack::Client::Auth->new($endpoint,
         'tenant'   => $ENV{'OS_TENANT_NAME'},
         'username' => $ENV{'OS_USERNAME'},
         'password' => $ENV{'OS_PASSWORD'}
@@ -49,7 +52,9 @@ OpenStack::Client - A cute little client to OpenStack services
     #
     use OpenStack::Client ();
 
-    my $glance = OpenStack::Client->new('http://glance.foo.bar:9292',
+    my $endpoint = 'http://glance.foo.bar:9292';
+
+    my $glance = OpenStack::Client->new($endpoint,
         'token' => {
             'id' => 'foo'
         }
@@ -97,7 +102,7 @@ A token obtained from a L<OpenStack::Client::Auth> object.
 sub new ($%) {
     my ($class, $endpoint, %opts) = @_;
 
-    die('No API endpoint provided') unless $endpoint;
+    die 'No API endpoint provided' unless $endpoint;
 
     $opts{'package_ua'}      ||= 'LWP::UserAgent';
     $opts{'package_request'} ||= 'HTTP::Request';
@@ -179,6 +184,10 @@ body.  This method may return the following:
 
 =back
 
+I<$body> may be supplied as a C<CODE> reference which, when called, will return
+a chunk of data to be supplied to the API endpoint.  The stream is ended when
+the supplied subroutine returns an empty string or undef.
+
 =item C<$client-E<gt>call(I<$method>, I<$headers>, I<$path>, I<$body>)>
 
 There exists a second form of C<call> that allows one to pass in
@@ -191,6 +200,10 @@ Headers are case I<insensitive>, and if one sets duplicate headers, one of
 them will get set; but there are no guarantess which one will. Repeat
 headers at your own risk.
 
+I<$body> may be supplied as a C<CODE> reference which, when called, will return
+a chunk of data to be supplied to the API endpoint.  The stream is ended when
+the supplied subroutine returns an empty string or undef.
+
 =over
 
 =item Accept
@@ -216,15 +229,20 @@ structure defined by the PATCH RFC (6902) governing "JavaScript Object
 Notation (JSON) Patch"; i.e., operations consisting of C<add>, C<replace>,
 or C<delete>.
 
-  my $headers  = { 'Content-Type' => 'application/openstack-images-v2.1-json-patch' };
-  my $response = $glance->call( q{PATCH}, $headers, qq[/v2/images/$image->{id}], \@image_updates )
+    my $headers  = {
+        'Content-Type' => 'application/openstack-images-v2.1-json-patch'
+    };
+
+    my $response = $glance->call('PATCH', $headers,
+        qq[/v2/images/$image->{id}], \@image_updates
+    );
 
 =back
 
-And except for C<x-auth-token>, any additional token will be added to the request.
+Except for C<X-Auth-Token>, any additional token will be added to the request.
 
 In exceptional conditions (such as when the service returns a 4xx or 5xx HTTP
-response), the client will C<die()> with the raw text response from the HTTP
+response), the client will die with the raw text response from the HTTP
 service, indicating the nature of the service-side failure to service the
 current call.
 
@@ -260,7 +278,19 @@ sub call {
         $request->header($name => $value);
     }
 
-    $request->content(JSON::encode_json($body)) if defined $body;
+    if (defined $body) {
+        #
+        # Allow the request body to be supplied by a subroutine reference
+        # which, when called, will supply a chunk of data returned as per the
+        # behavior of LWP::UserAgent.  This is useful for uploading arbitrary
+        # amounts of data in a request body.
+        #
+        if (ref($body) =~ /CODE/) {
+            $request->content($body);
+        } else {
+            $request->content(JSON::encode_json($body));
+        }
+    }
 
     my $response = $self->{'ua'}->request($request);
 
@@ -270,7 +300,7 @@ sub call {
     if ($response->code =~ /^[45]\d{2}$/) {
         $content ||= "@{[$response->code]} Unknown error";
 
-        die($content);
+        die $content;
     }
 
     if (lc($type) =~ qr{^application/json}i && defined $content && length $content) {
@@ -280,35 +310,53 @@ sub call {
     }
 }
 
-# internal method for call() to process headers; returns a list of hash references - one for each header/value
+sub _lc_merge {
+    my ($a, $b, %opts) = @_;
+
+    my %lc_keys_a = map {
+        lc $_ => $_
+    } keys %{$a};
+
+    foreach my $key_b (keys %{$b}) {
+        my $key_a = $lc_keys_a{lc $key_b};
+
+        if (!defined($key_a)) {
+            $a->{$key_b} = $b->{$key_b};
+        } elsif (exists $a->{$key_a} && $opts{'replace'}) {
+            $a->{$key_a} = $b->{$key_b};
+        }
+    }
+
+    return;
+}
+
+#
+# Internal method for call() to process headers; returns a list of header name
+# and value pairs
+#
 sub _get_headers_list {
     my ($self, $headers) = @_;
 
-    # place to store supplied headers with lowercase keys 
-    my $lc_headers = {};
-
-    # lowercase supplied headers so we know what to add in place of our default headers lc key collisions go to last set 
-    foreach my $header (keys %{$headers}) {
-        my $lc_header = lc $header;
-        $lc_headers->{$lc_header} = $headers->{$header};
-    }
-
-    # default set of headers, set defaults if not specified explicitly
-    my @headers = (
-        'Accept'          => $lc_headers->{'accept'}          // 'application/json, text/plain',
-        'Accept-Encoding' => $lc_headers->{'accept-encoding'} // 'identity, gzip, deflate, compress',
-        'Content-Type'    => $lc_headers->{'content-type'}    // 'application/json'
+    my %DEFAULTS = (
+        'Accept'          => 'application/json, text/plain',
+        'Accept-Encoding' => 'identity, gzip, deflate, compress',
+        'Content-Type'    => 'application/json'
     );
 
-    # add all we don't care about 
-    foreach my $header (grep( !/^Accept$|^Accept\-Encoding$|^Content\-Type$|^X-Auth-Token$/i, keys %{$lc_headers})) {
-        push @headers, $header => $lc_headers->{$header};
-    }
+    #
+    # The client should be not adding X-Auth-Token explicitly, so force it to
+    # the one received during authentication
+    #
+    my %OVERRIDES = (
+        'X-Auth-Token' => $self->{'token'}->{'id'}
+    );
 
-    # client should be not adding X-Auth-Token explicitly, so force it to the one received during authentication
-    push @headers, ( 'X-Auth-Token' => $self->{'token'}->{'id'} ) if defined $self->{'token'}->{'id'};
+    my %new_headers = %{$headers};
 
-    return @headers;
+    _lc_merge(\%new_headers, \%DEFAULTS);
+    _lc_merge(\%new_headers, \%OVERRIDES, 'replace' => 1);
+
+    return %new_headers;
 }
 
 =back
@@ -377,7 +425,7 @@ sub each ($$@) {
     } elsif (scalar @args == 1) {
         ($callback) = @args;
     } else {
-        die('Invalid number of arguments');
+        die 'Invalid number of arguments';
     }
 
     while (defined $path) {
@@ -414,14 +462,14 @@ sub every ($$$@) {
     } elsif (scalar @args == 1) {
         ($callback) = @args;
     } else {
-        die('Invalid number of arguments');
+        die 'Invalid number of arguments';
     }
 
     while (defined $path) {
         my $result = $self->get($path, %{$opts});
 
         unless (defined $result->{$attribute}) {
-            die("Response from $path does not contain attribute '$attribute'");
+            die "Response from $path does not contain attribute '$attribute'";
         }
 
         foreach my $item (@{$result->{$attribute}}) {
